class LinearMap____:
    pass


class Matrix____:
    pass


class Span____:
    pass


class Vector____:
    pass


class Complex____:
    pass


class Field____:
    pass


class VectorSpace____:
    pass


class PolynomialSimple____:
    pass
